'''
Relativictic Difference Scheme Particle-in-Cell code (REDPIC) solver file.

'''

import numpy as np
import pandas as pd
from numba import jit, prange
from multiprocessing import Process
from scipy import misc
from .constants import *
from .beam import *
from .accelerator import *

__all__ = [ 'Simulation',
            'Sim'
             ]

def get_field_accelerator(acc: Accelerator, type: str,
                          x: np.array, y: np.array, z: np.array) -> (np.array,
                                                                     np.array,
                                                                     np.array):
    ''' Get an electric or magnetic field at a specific location in the accelerator

    '''
    assert type == 'E' or type == 'B', 'Check type field! (E or B)'

    dz = acc.dz
    offset_x = acc.Dx(z)
    offset_xp = acc.Dxp(z)
    offset_y = acc.Dy(z)
    offset_yp = acc.Dyp(z)
    x_corr = x - offset_x
    y_corr = y - offset_y
    r_corr = np.sqrt(x_corr*x_corr + y_corr*y_corr)

    if type == 'E':
        Ez = acc.Ez(z)
        dEzdz = acc.dEzdz(z)
        d2Ezdz2 = misc.derivative(acc.dEzdz, z, dx=dz, n=1)
        Ez = Ez - d2Ezdz2*r_corr*r_corr/4         # row remainder
        Ex = - dEzdz*x_corr/2 + Ez*offset_xp                # row remainder
        Ey = - dEzdz*y_corr/2 + Ez*offset_yp                # row remainder
        return Ex, Ey, Ez
    if type == 'B':
        Bx = acc.Bx(z)
        By = acc.By(z)
        Bz = acc.Bz(z)
        dBzdz = acc.dBzdz(z)
        d2Bzdz2 = misc.derivative(acc.dBzdz, z, dx=dz, n=1)
        Gz = acc.Gz(z)
        Bz = Bz - d2Bzdz2*r_corr*r_corr/4              # row remainder
        Bx = Bx + Gz*y_corr - dBzdz*x_corr/2 + Bz*offset_xp     # row remainder
        By = By + Gz*x_corr - dBzdz*y_corr/2 + Bz*offset_yp     # row remainder
        return Bx, By, Bz

@jit(nopython=True, parallel=True, fastmath=True, cache=True, nogil=True)
def sum_field_particles(x: np.array, y: np.array, z: np.array,
                        z_start: float, z_stop: float) -> (np.array, np.array, np.array):
    ''' Sum particles fields

    '''
    n = int(len(x))
    Fx = np.zeros(n)
    Fy = np.zeros(n)
    Fz = np.zeros(n)
    for i in prange(n):
        if z[i] >= z_start and z[i] <= z_stop:
            for j in prange(n):
                if z[j] >= z_start and z[j] <= z_stop and i != j:
                    Fx[i] += (x[i]-x[j]) / ((x[j] - x[i])**2 + (y[j] - y[i])**2 + (z[j] - z[i])**2)**(3/2)
                    Fy[i] += (y[i]-y[j]) / ((x[j] - x[i])**2 + (y[j] - y[i])**2 + (z[j] - z[i])**2)**(3/2)
                    Fz[i] += (z[i]-z[j]) / ((x[j] - x[i])**2 + (y[j] - y[i])**2 + (z[j] - z[i])**2)**(3/2)
    return Fx, Fy, Fz

def get_field_beam(beam: Beam, acc: Accelerator, type: str,
                   x: np.array, y: np.array, z: np.array) -> (np.array, np.array, np.array):
    ''' Get an electric [MV/m] or magnetic [T] field space charge at a specific location in the accelerator

    '''
    assert type == 'E' or type == 'B', 'Check type field! (E or B)'
    Q = beam.charge
    q = beam.charge / beam.n
    if type == 'E':
        Ex, Ey, Ez = sum_field_particles(x, y, z, acc.z_start, acc.z_stop)
        return ke/(4*np.pi)*q*Ex/1e6, ke/(4*np.pi)*q*Ey/1e6, ke/(4*np.pi)*q*Ez/1e6
    if type == 'B':
        Bx, By, Bz = sum_field_particles(x, y, z, acc.z_start, acc.z_stop)
        return km/(4*np.pi)*q*Bx, -km/(4*np.pi)*q*By, 0*km/(4*np.pi)*q*Bz

@jit(nopython=True, parallel=True, fastmath=True, cache=True, nogil=True)
def first_step_red(x: np.array, y: np.array, z: np.array,
                   px: np.array, py: np.array, pz: np.array,
                   Fx: np.array, Fy: np.array, Fz: np.array,
                   z_start: float, z_stop: float) -> (np.array, np.array, np.array,
                                                      np.array, np.array, np.array):
    ''' First step for Relativictic Difference Scheme

    '''
    n = int(len(x))
    for i in prange(n):
        if z[i] >= z_start and z[i] <= z_stop:
            px[i] += 2*Fx[i]
            py[i] += 2*Fy[i]
            pz[i] += 2*Fz[i]
            gamma = (1 + px[i]**2 + py[i]**2 + pz[i]**2)**(1/2)
            vx = px[i]/gamma
            vy = py[i]/gamma
            vz = pz[i]/gamma
            x[i] += vx
            y[i] += vy
            z[i] += vz
        else:
            gamma = (1 + px[i]**2 + py[i]**2 + pz[i]**2)**(1/2)
            vz = pz[i]/gamma
            z[i] += vz
    return x, y, z, px, py, pz

@jit(nopython=True, parallel=True, fastmath=True, cache=True, nogil=True)
def second_step_red(x: np.array, y: np.array, z: np.array,
                    px: np.array, py: np.array, pz: np.array,
                    Fx: np.array, Fy: np.array, Fz: np.array,
                    z_start: float, z_stop: float) -> (np.array, np.array, np.array,
                                                       np.array, np.array, np.array):
    ''' Second step for Relativictic Difference Scheme

    '''
    n = int(len(x))
    for i in prange(n):
        if z[i] >= z_start and z[i] <= z_stop:
            gamma = (1 + px[i]**2 + py[i]**2 + pz[i]**2)**(1/2)
            vx = px[i]/gamma
            vy = py[i]/gamma
            vz = pz[i]/gamma
            b2 = 1 + Fx[i]**2 + Fy[i]**2 + Fz[i]**2
            b1 = 2 - b2
            b3 = 2 * (vx*Fx[i] + vy*Fy[i] + vz*Fz[i])
            fx = 2 * (vy*Fz[i] - vz*Fy[i])
            fy = 2 * (vz*Fx[i] - vx*Fz[i])
            fz = 2 * (vx*Fy[i] - vy*Fx[i])
            vx = (vx*b1 + fx + Fx[i]*b3)/b2
            vy = (vy*b1 + fy + Fy[i]*b3)/b2
            vz = (vz*b1 + fz + Fz[i]*b3)/b2
            x[i] += vx
            y[i] += vy
            z[i] += vz
            px[i] = vx*gamma
            py[i] = vy*gamma
            pz[i] = vz*gamma
        else:
            gamma = (1 + px[i]**2 + py[i]**2 + pz[i]**2)**(1/2)
            vz = pz[i]/gamma
            z[i] += vz
    return x, y, z, px, py, pz

class Simulation:
    ''' Simulation of the beam tracking in the accelerator

    '''
    def __init__(self, beam, accelerator):
        self.beam = beam
        self.acc = accelerator

    def track(self, *, n_files: int=20, path: str='') -> None:
        ''' Tracking!

        '''
        assert n_files > 0, 'The number of files (n_files) must be a positive number!'

        # Init parameterss
        Y = self.beam.da
        Y[2] = Y[2] + self.acc.z_start - max(Y[2]) # set initial beam position

        z_start = self.acc.z_start
        z_stop = self.acc.z_stop
        dz = self.acc.dz
        dt = dz / c
        len_beam = max(Y[2]) - min(Y[2])
        t_max = (z_stop - z_start) / c

        m = self.beam.type.mass
        q = self.beam.type.charge

        P0 = m*c*c / (e*1e6)
        E0 = m*c / (q*dt*1e6)
        B0 = m / (q*dt)

        # RED
        for t in np.arange(0, t_max, 2*dt):

            # into the internal system
            Y[0], Y[1], Y[2] = Y[0]/dz, Y[1]/dz, Y[2]/dz
            Y[3], Y[4], Y[5] = Y[3]/P0, Y[4]/P0, Y[5]/P0

            # get electric field from accelerator
            Ex, Ey, Ez = get_field_accelerator(self.acc, 'E',
                                               Y[0]*dz, Y[1]*dz, Y[2]*dz)
            Ex, Ey, Ez = Ex/E0, Ey/E0, Ez/E0

            # get electric field from beam
            if self.beam.charge:
                ex, ey, ez = get_field_beam(self.beam, self.acc, 'E',
                                            Y[0]*dz, Y[1]*dz, Y[2]*dz)
                ex, ey, ez = ex/E0, ey/E0, ez/E0
                Ex, Ey, Ez = Ex + ex, Ey + ey, Ez + ez

            # first step RED
            Y[0], Y[1], Y[2], Y[3], Y[4], Y[5] = first_step_red(Y[0], Y[1], Y[2],
                                                                Y[3], Y[4], Y[5],
                                                                Ex, Ey, Ez,
                                                                z_start/dz, z_stop/dz)
            gamma = np.sqrt(1 + Y[3]*Y[3] + Y[4]*Y[4] + Y[5]*Y[5])
            vx, vy, vz = Y[3]/gamma, Y[4]/gamma, Y[5]/gamma

            # get magnetic field from accelerator
            Bx, By, Bz = get_field_accelerator(self.acc, 'B',
                                               Y[0]*dz, Y[1]*dz, Y[2]*dz)
            Bx, By, Bz = Bx/B0/gamma, By/B0/gamma, Bz/B0/gamma

            # get magnetic field from beam
            if self.beam.charge:
                bx, by, bz = get_field_beam(self.beam, self.acc, 'B',
                                            Y[0]*dz, Y[1]*dz, Y[2]*dz)
                bx, by, bz = bx*vz/B0, by*vz/B0, bz*vz/B0
                Bx, By, Bz = Bx + bx, By + by, Bz + bz

            # second step RED
            Y[0], Y[1], Y[2], Y[3], Y[4], Y[5] = second_step_red(Y[0], Y[1], Y[2],
                                                                 Y[3], Y[4], Y[5],
                                                                 Bx, By, Bz,
                                                                 z_start/dz, z_stop/dz)
            gamma = np.sqrt(1 + Y[3]*Y[3] + Y[4]*Y[4] + Y[5]*Y[5])
            vx, vy, vz = Y[3]/gamma, Y[4]/gamma, Y[5]/gamma

            # into the SI
            Y[0], Y[1], Y[2] = Y[0]*dz, Y[1]*dz, Y[2]*dz
            Y[3], Y[4], Y[5] = Y[3]*P0, Y[4]*P0, Y[5]*P0
            Bx, By, Bz = Bx*gamma*B0, By*gamma*B0, Bz*gamma*B0
            Ex, Ey, Ez = Ex*E0, Ey*E0, Ez*E0

            # output in files (symbolBeam.XXXX.csv)
            progress = t / t_max * 100
            meters = z_start + t * c
            X = np.row_stack((Y[0], Y[1], Y[2], Y[3], Y[4], Y[5],
                              Bx, By, Bz, Ex, Ey, Ez))
            Xt = np.transpose(X)
            df = pd.DataFrame(Xt, columns=[ 'x','y','z','px','py','pz',
                             'Bx', 'By', 'Bz', 'Ex', 'Ey', 'Ez' ])
            file_name = path + self.beam.type.symbol + 'Beam.'+ '%04.0f' % (meters * 100) +'.csv'
            if progress % (100 // n_files) < 2*dt / t_max * 100:
                writer = Process(target=df.to_csv, args=(file_name,), daemon=True)
                writer.start()
            print( '\rz = %.2f m (%.1f %%) ' % (meters, progress), end='')

Sim = Simulation




import redpic as rp
import kenv as kv
import pandas as pd
import numpy as np
from bokeh.models import (Panel, PreText, RadioButtonGroup, Select,
 						  FileInput, Button, TextInput, ColumnDataSource,
						  RangeSlider, Div)
from bokeh.plotting import figure
from bokeh.layouts import column, row
from bokeh.models.widgets import Slider
from os.path import dirname, join

def tuning_tab(beam, acc):
    '''Tuning accelerator

	'''
    pre = PreText(text='''You can configure accelerator elements.''')
    element_button = RadioButtonGroup(
        labels=['Accels', 'Solenoids', 'Quadrupoles'], active=0)
    tune_button = Button(label='Tune', button_type="success")
    select = Select(options=[itm.name for itm in acc.Ez_beamline.values()])
    select_maxfield = Slider(start=-5.0, end=5.0,
				    		 step=0.1, value=0.0,
							 title='MaxField [MV/m]', format='0[.]0')

    line_source = ColumnDataSource(data={'z': [0, 0], 'x': [-0.1, 0.1]})
    field_source = ColumnDataSource(data={'z': acc.z, 'Fz': acc.Ez(acc.z)})
    field_plot = figure(x_axis_label='z [m]', y_axis_label='Ez [MV/m]',
                        width=330, height=210)
    field_plot.line('z', 'Fz', source=field_source, line_color='#3A5785',
                    line_alpha=0.7, line_width=1)

    envelope_plot = figure(x_axis_label='z [m]', y_axis_label='Envelope, x [m]',
                        width=650, height=250)
    envelope_source = ColumnDataSource(data={'z': acc.z, 'x': 0*acc.z,
                                             '-x': 0*acc.z})
    centroid_source = ColumnDataSource(data={'z': acc.z, 'x': 0*acc.z})
    envelope_plot.line('z', 'x', source=envelope_source, line_color='#3A5785',
                        line_alpha=0.7, line_width=1)
    envelope_plot.line('z', '-x', source=envelope_source, line_color='#3A5785',
                        line_alpha=0.7, line_width=1)
    envelope_plot.line('z', 'x', source=line_source, line_color='#3A5785',
                        line_alpha=0.1, line_width=10)
    envelope_plot.line('z', 'x', source=centroid_source, line_color='#3A5785',
                        line_alpha=0.7, line_width=1, line_dash=[3,5])

    controls = row(column(pre, element_button, select, select_maxfield, tune_button),
                   field_plot)
    tab = Panel(child=column(controls, envelope_plot), title='Tuning')

    def element_handler(new, acc=acc):
        if new == 0:
            select.options=[itm.name for itm in acc.Ez_beamline.values()]
            select_maxfield.start = -5.0
            select_maxfield.end = 5.0
            select_maxfield.step = 0.1
            if select.options == []:
                select_maxfield.value = 0.0
                select.value = ''
            else:
                select_maxfield.value = acc.Ez_beamline[select.options[0]].max_field
                select.value = [itm.name for itm in acc.Ez_beamline.values()][0]
            select_maxfield.format='0[.]0'
            select_maxfield.title = 'MaxField [MV/m]'
            field_source.data = {'z': acc.z, 'Fz': acc.Ez(acc.z)}
            field_plot.yaxis.axis_label = 'Ez [MV/m]'
        if new == 1:
            select.options=[itm.name for itm in acc.Bz_beamline.values()]
            select_maxfield.start = 0.000
            select_maxfield.end = 0.500
            select_maxfield.step = 0.001
            if select.options == []:
                select_maxfield.value = 0.000
                select.value = ''
            else:
                select_maxfield.value = acc.Bz_beamline[select.options[0]].max_field
                select.value = [itm.name for itm in acc.Bz_beamline.values()][0]
            select_maxfield.format='0[.]000'
            select_maxfield.title = 'MaxField [T]'
            field_source.data = {'z': acc.z, 'Fz': acc.Bz(acc.z)}
            field_plot.yaxis.axis_label = 'Bz [T]'
        if new == 2:
            select.options=[itm.name for itm in acc.Gz_beamline.values()]
            select_maxfield.start = -10.0
            select_maxfield.end = 10.0
            select_maxfield.step = 0.1
            if select.options == []:
                select_maxfield.value = 0.0
                select.value = ''
            else:
                select_maxfield.value = acc.Gz_beamline[select.options[0]].max_field
                select.value = [itm.name for itm in acc.Gz_beamline.values()][0]
            select_maxfield.format='0[.]0'
            select_maxfield.title = 'MaxField [T/m]'
            field_source.data = {'z': acc.z, 'Fz': acc.Gz(acc.z)}
            field_plot.yaxis.axis_label = 'Gz [T/m]'

    def update_select_maxfield(attrname, old, new, acc=acc):
        if new != '':
            if element_button.active == 0:
                select_maxfield.value = acc.Ez_beamline[new].max_field
                line_source.data = {'z': [acc.Ez_beamline[new].z0, acc.Ez_beamline[new].z0],
                                    'x': [-0.1, 0.1]}
            if element_button.active == 1:
                select_maxfield.value = acc.Bz_beamline[new].max_field
                line_source.data = {'z': [acc.Bz_beamline[new].z0, acc.Bz_beamline[new].z0],
                                    'x': [-0.1, 0.1]}
            if element_button.active == 2:
                select_maxfield.value = acc.Gz_beamline[new].max_field
                line_source.data = {'z': [acc.Gz_beamline[new].z0, acc.Gz_beamline[new].z0],
                                    'x': [-0.1, 0.1]}

    def tune_handler(new, beam=beam, acc=acc):
        if select.value != '':
            if element_button.active == 0:
                acc.Ez_beamline[select.value].max_field = select_maxfield.value
                acc.compile()
                field_source.data = {'z': acc.z, 'Fz': acc.Ez(acc.z)}
            if element_button.active == 1:
                acc.Bz_beamline[select.value].max_field = select_maxfield.value
                acc.compile()
                field_source.data = {'z': acc.z, 'Fz': acc.Bz(acc.z)}
            if element_button.active == 2:
                acc.Gz_beamline[select.value].max_field = select_maxfield.value
                acc.compile()
                field_source.data = {'z': acc.z, 'Fz': acc.Gz(acc.z)}

        beam.df['p2'] = beam.df['pz']*beam.df['pz'] + beam.df['px']*beam.df['px'] + beam.df['py']*beam.df['py']
        E = beam.df['p2'].mean()**0.5 - beam.type.mass*rp.c**2/rp.e/1e6
        I = np.sign(beam.type.charge) * beam.charge * rp.c / (beam.df['z'].max()-beam.df['z'].min())
        X = beam.df['x'].max()
        Y = beam.df['y'].max()
        Xp = beam.df['px'].max() / beam.df['pz'].max()
        Yp = beam.df['py'].max() / beam.df['pz'].max()
        X_off = beam.df['x'].mean()
        Y_off = beam.df['x'].mean()
        Xp_off = beam.df['px'].mean() / beam.df['pz'].max()
        Yp_off = beam.df['py'].mean() / beam.df['pz'].max()
        beta = beam.df['pz'].max() / (E + beam.type.mass*rp.c**2/rp.e/1e6)
        gamma = 1/(1-beta**2)**0.5
        df = beam.df
        df['xp'] = df['px']/df['pz']
        df['yp'] = df['py']/df['pz']
        df['xp2'] = df['xp']*df['xp']
        df['yp2'] = df['yp']*df['yp']
        df['x xp'] = df['x']*df['xp']
        df['y yp'] = df['y']*df['yp']
        df['x2'] = df['x']*df['x']
        df['y2'] = df['y']*df['y']
        df['rms x'] = df['x2'].mean()**0.5
        df['rms y'] = df['y2'].mean()**0.5
        df['rms x Emittance'] = (df['x2'].mean()*df['xp2'].mean() - df['x xp'].mean()**2)**0.5
        df['rms y Emittance'] = (df['y2'].mean()*df['yp2'].mean() - df['y yp'].mean()**2)**0.5
        Emitt_x = df['rms x Emittance'].max()*gamma*beta
        Emitt_y = df['rms y Emittance'].max()*gamma*beta
        kv_beam = kv.Beam(energy=E,
                         current=I,
                         radius_x=X,
                         radius_y=Y,
                         radius_xp=Xp,
                         radius_yp=Yp,
                         normalized_emittance_x=Emitt_x,
                         normalized_emittance_y=Emitt_y,
                         x=X_off, y=Y_off,
                         xp=Xp_off, yp=Yp_off,
                         charge=np.sign(beam.type.charge))
        kv_sim = kv.Simulation(kv_beam, acc)
        kv_sim.track()
        envelope_source.data = {'z': acc.z, 'x': kv_sim.envelope_x(acc.z),
                                '-x': -kv_sim.envelope_x(acc.z)}
        centroid_source.data = {'z': acc.z, 'x': kv_sim.centroid_x(acc.z)}

    element_button.on_click(element_handler)
    select.on_change('value', update_select_maxfield)
    tune_button.on_click(tune_handler)

    return tab




import redpic as rp
import kenv as kv
import pandas as pd
import numpy as np
from bokeh.models import (Panel, PreText, RadioButtonGroup, Select,
                           FileInput, Button, TextInput, ColumnDataSource,
                          RangeSlider)
from bokeh.plotting import figure
from bokeh.layouts import column, row
from bokeh.models.widgets import Slider
from os.path import dirname, join


def accelerator_tab(acc):
    '''Creating an accelerator with parameters

    '''

    length = RangeSlider(start=0, end=25, value=(0,1), step=.1,
                         title='Length [m]', format='0[.]0')
    pre = PreText(text='''Here you can compile an accelerator.''')
    compile_button = Button(label='Compile', button_type="success", height=45)
    element_button = RadioButtonGroup(
        labels=['Accels', 'Solenoids', 'Quadrupoles'], active=0)

    name_input = TextInput(value="Acc. ", title="Name:", width=100)
    z_input = TextInput(value="0", title="z [m]:", width=100)
    field_input = TextInput(value="0", title="MaxField [MV/m]:", width=100)
    file_input = FileInput(accept='.csv, .dat, .txt')
    parameters_input = column(file_input,
                              row(name_input, z_input, field_input))

    source = ColumnDataSource(data={'z': acc.z, 'Fz': acc.Ez(acc.z)})
    field_plot = figure(x_axis_label='z [m]', y_axis_label='Ez [MV/m]',
                        width=650, height=250)
    field_plot.line('z', 'Fz', source=source, line_color='#3A5785',
                    line_alpha=0.7, line_width=1)

    controls = row(column(pre, length, compile_button),
                   column(element_button, parameters_input))

    # Create a row layout
    layout = column(controls, field_plot)
    tab = Panel(child=layout, title='Accelerator')

    def compile_handler(new, acc=acc):
        (acc.z_start, acc.z_stop) = length.value
        acc.z = acc.parameter = np.arange(acc.z_start, acc.z_stop, acc.dz)
        name = name_input.value
        file_name = dirname(__file__) + '/data/' + file_input.filename
        position = float(z_input.value)
        max_field = float(field_input.value)
        if element_button.active == 0:
            acc.add_accel(name, position, max_field, file_name)
            acc.compile()
            source.data = {'z': acc.z, 'Fz': acc.Ez(acc.z)}
            field_input.title = 'MaxField [MV/m]:'
            field_plot.yaxis.axis_label = 'Ez [MV/m]'
        if element_button.active == 1:
            acc.add_solenoid(name, position, max_field, file_name)
            acc.compile()
            source.data = {'z': acc.z, 'Fz': acc.Bz(acc.z)}
            field_input.title = 'MaxField [T]:'
            field_plot.yaxis.axis_label = 'Bz [T]'
        if element_button.active == 2:
            acc.add_quad(name, position, max_field, file_name)
            acc.compile()
            source.data = {'z': acc.z, 'Fz': acc.Gz(acc.z)}
            field_input.title = 'MaxField [T/m]:'
            field_plot.yaxis.axis_label = 'Gz [T/m]'
        print(acc)

    def element_handler(new):
        if new == 0:
            source.data = {'z': acc.z, 'Fz': acc.Ez(acc.z)}
            field_input.title = 'MaxField [MV/m]:'
            name_input.value = 'Acc. '
            field_plot.yaxis.axis_label = 'Ez [MV/m]'
        if new == 1:
            source.data = {'z': acc.z, 'Fz': acc.Bz(acc.z)}
            field_input.title = 'MaxField [T]:'
            name_input.value = 'Sol. '
            field_plot.yaxis.axis_label = 'Bz [T]'
        if new == 2:
            source.data = {'z': acc.z, 'Fz': acc.Gz(acc.z)}
            field_input.title = 'MaxField [T/m]:'
            name_input.value = 'Quad. '
            field_plot.yaxis.axis_label = 'Gz [T/m]'

    compile_button.on_click(compile_handler)
    element_button.on_click(element_handler)

    return tab



# Library
import redpic as rp
import kenv as kv
import numpy as np
import pandas as pd
import glob

# Bokeh basics
from bokeh.io import curdoc
from bokeh.models.widgets import Tabs
from bokeh.models import ColumnDataSource
from os.path import dirname, join

# Each tab is drawn by one script
from beam import beam_tab
from accelerator import accelerator_tab
from tuning import tuning_tab
from production import production_tab

print(rp.__doc__)
print('Version: ' + rp.__version__)
# Initial parameters
beam = rp.Beam(rp.electron)
beam.generate(rp.Distribution('KV', x=1, y=1, z=1, px=1, py=1, pz=1), n=1)
accelerator = rp.Accelerator(0, 1, 0.001)
accelerator.compile()

# Create each of the tabs
tab1 = beam_tab(beam)
tab2 = accelerator_tab(accelerator)
tab3 = tuning_tab(beam, accelerator)
tab4 = production_tab(beam, accelerator)

# Put all the tabs into one application
tabs = Tabs(tabs=[tab1, tab2, tab3, tab4])

# Put the tabs in the current document for display
curdoc().add_root(tabs)
curdoc().title = "REDPIC"



import redpic as rp
import kenv as kv
import pandas as pd
import numpy as np
import glob
from multiprocessing import Process
from bokeh.models import (Panel, PreText, RadioButtonGroup, Select,
                           FileInput, Button, TextInput, ColumnDataSource,
                          RangeSlider, Div)
from bokeh.plotting import figure, curdoc
from bokeh.layouts import column, row
from bokeh.models.widgets import Slider
from os.path import dirname, join


def simulation_tab(beam, acc):
    '''Simulation

    '''
    pre = PreText(text='''Here you can run the beam simulation in the accelerator.''')

    start_button = Button(label='Start', button_type="success")

    refresh_button = Button(label='Refresh', button_type="warning")

    p = figure(x_axis_label='z [m]', y_axis_label='x [m]', x_range=(acc.z_start, acc.z_stop),
                        width=600, height=210)
    source = ColumnDataSource(data={'x': beam.df['x'], 'z': beam.df['z']})
    p.scatter('z', 'x', source=source, size=0.5, color="#3A5785", alpha=0.5)

    controls = row(start_button, refresh_button)
    tab = Panel(child=column(p, controls), title='Simulation')

    def start_handler(new, beam=beam, acc=acc):
        simulation = rp.Simulation(beam, acc)
        simulation.track(n_files=30, path=dirname(__file__) + '/data/')

    def play_handler(beam=beam, acc=acc):
        track_files = np.sort(glob.glob(dirname(__file__) + '/data/'+ '*.*[0-9].csv'))
        if len(track_files) > 0:
            df = pd.read_csv(track_files[-1], dtype='float32')
            source.data = {'x': df['x'], 'z': df['z']}

    start_button.on_click(start_handler)
    play_button.on_click(play_handler)

    return tab



import redpic as rp
import kenv as kv
import pandas as pd
import numpy as np
import glob
from scipy import interpolate
from multiprocessing import Process
from bokeh.models import (Panel, PreText, RadioButtonGroup, RadioGroup, Select,
                          FileInput, Button, TextInput, ColumnDataSource,
                          RangeSlider, Div)
from bokeh.plotting import figure, curdoc
from bokeh.layouts import column, row
from bokeh.models.widgets import Slider, Tabs
from bokeh.models.ranges import Range1d
from os.path import dirname, join

def window_rms(a: np.array, window_size: int=1_000):
    '''windows RMS calculation

    '''
    a2 = np.power(a, 2)
    window = np.ones(int(window_size)) / float(window_size)
    return np.sqrt(np.convolve(a2, window, 'same'))

def window_mean(a: np.array, window_size: int=1_000):
    '''windows MEAN calculation

    '''
    a2 = a
    window = np.ones(int(window_size)) / float(window_size)
    return np.convolve(a2, window, 'same')

def production_tab(beam, acc):
    '''Simulation

    '''
    pre = PreText(text='''Track beam in the accelerator.''')

    track_button = Button(label='Track', button_type="success", width=315)

    refresh_button = Button(label='Refresh', button_type="warning", width=315)

    mode_button = RadioButtonGroup(
        labels=['Phase', 'Line', 'Field'], active=0)

    phase_button = RadioButtonGroup(
        labels=['z-x', 'z-y', 'x-y', 'x-px', 'x-py', 'y-px', 'y-py', 'px-py'],
                active=0)
    field_button = RadioButtonGroup(
        labels=['z-Ez', 'z-Ex', 'z-Ey', 'z-Bz', 'z-Bx', 'z-By'], active=0)
    line_button = RadioButtonGroup(
        labels=['rms x', 'rms y', 'rms x Emitt', 'rms y Emitt', 'avg Energy',
                'x β-func', 'y β-func', 'avg x', 'avg y'], active=0)

    file_input = FileInput(accept='.csv')

    phase_source = ColumnDataSource(data={'x': beam.df['z'], 'y': beam.df['x']})
    phase_plot = figure(x_axis_label='z [m]', y_axis_label='x [m]',
                        width=650, height=250)
    phase_plot.scatter('x', 'y', source=phase_source, size=0.5, color="#3A5785",
                        alpha=0.5)

    beam.df['x2'] = beam.df['x']*beam.df['x']
    beam.df['rms x'] = beam.df['x2'].mean()**0.5
    line_source = ColumnDataSource(data={'x': beam.df['z'], 'y': beam.df['rms x']})
    line_plot = figure(x_axis_label='z [m]', y_axis_label='rms x [m]',
                       width=650, height=250)
    line_plot.line('x', 'y', source=line_source, line_color='#3A5785',
                   line_alpha=0.7, line_width=1)

    field_source = ColumnDataSource(data={'x': beam.df['z'], 'y': beam.df['rms x']})
    field_plot = figure(x_axis_label='z [m]', y_axis_label='Ez [MV/m]',
                        width=650, height=250)
    field_plot.line('x', 'y', source=field_source, line_color='#3A5785',
                    line_alpha=0.7, line_width=1)

    buttons = row(track_button, refresh_button)
    controls = row(pre, file_input)
    phase_tab = Panel(child=column(phase_button, phase_plot), title='Phase')
    line_tab = Panel(child=column(line_button, line_plot), title='Line')
    field_tab = Panel(child=column(field_button, field_plot), title='Field')
    tabs = Tabs(tabs=[phase_tab, line_tab, field_tab])
    tab = Panel(child=column(controls, tabs, buttons), title='Production')

    def calculate(df):
        df = df.sort_values('z')
        df['avg p'] = window_mean((df['px']*df['px'] + df['py']*df['py'] + df['pz']*df['pz'])**0.5)
        df['xp'] = df['px']/df['pz']
        df['yp'] = df['py']/df['pz']
        df['x xp'] = window_mean(df['x']*df['xp'])
        df['y yp'] = window_mean(df['y']*df['yp'])
        df['rms x'] = window_rms(df['x'])
        df['rms y'] = window_rms(df['y'])
        df['rms xp'] = window_rms(df['xp'])
        df['rms yp'] = window_rms(df['yp'])
        df['rms x Emittance'] = (df['rms x']**2 * df['rms xp']**2 - df['x xp']**2)**0.5
        df['rms y Emittance'] = (df['rms y']**2 * df['rms yp']**2 - df['y yp']**2)**0.5
        df['avg Energy'] = df['avg p'] - beam.type.mass*rp.c**2/rp.e/1e6
        df['x β-function'] = df['rms x']**2/df['rms x Emittance']
        df['y β-function'] = df['rms y']**2/df['rms y Emittance']
        df['centroid x'] = window_mean(df['x'])
        df['centroid y'] = window_mean(df['y'])
        df['avg Ez'] = window_mean(df['Ez'])
        df['avg Ex'] = window_mean(df['Ex'])
        df['avg Ey'] = window_mean(df['Ey'])
        df['avg Bz'] = window_mean(df['Bz'])
        df['avg Bx'] = window_mean(df['Bx'])
        df['avg By'] = window_mean(df['By'])
        return df

    def track_handler(new, beam=beam, acc=acc):
        simulation = rp.Simulation(beam, acc)
        simulation.track(n_files=30, path=dirname(__file__) + '/data/')

    def refresh_handler(beam=beam, acc=acc):
        fname = dirname(__file__) + '/data/' + file_input.filename
        df = pd.read_csv(fname, dtype='float32')
        df = calculate(df)
        df = df[df.z >= acc.z_start]
        df = df[df.z <= acc.z_stop]
        if tabs.active == 0:
            if phase_button.active == 0:
                phase_plot.xaxis.axis_label = 'z [m]'
                phase_plot.yaxis.axis_label = 'x [m]'
                phase_source.data = {'x': df['z'], 'y': df['x']}
            if phase_button.active == 1:
                phase_plot.xaxis.axis_label = 'z [m]'
                phase_plot.yaxis.axis_label = 'y [m]'
                phase_source.data = {'x': df['z'], 'y': df['y']}
            if phase_button.active == 2:
                phase_plot.xaxis.axis_label = 'x [m]'
                phase_plot.yaxis.axis_label = 'y [m]'
                phase_source.data = {'x': df['x'], 'y': df['y']}
            if phase_button.active == 3:
                phase_plot.xaxis.axis_label = 'x [m]'
                phase_plot.yaxis.axis_label = 'px [MeV/c]'
                phase_source.data = {'x': df['x'], 'y': df['px']}
            if phase_button.active == 4:
                phase_plot.xaxis.axis_label = 'x [m]'
                phase_plot.yaxis.axis_label = 'py [MeV/c]'
                phase_source.data = {'x': df['x'], 'y': df['py']}
            if phase_button.active == 5:
                phase_plot.xaxis.axis_label = 'y [m]'
                phase_plot.yaxis.axis_label = 'px [MeV/c]'
                phase_source.data = {'x': df['y'], 'y': df['px']}
            if phase_button.active == 6:
                phase_plot.xaxis.axis_label = 'y [m]'
                phase_plot.yaxis.axis_label = 'py [MeV/c]'
                phase_source.data = {'x': df['y'], 'y': df['py']}
            if phase_button.active == 7:
                phase_plot.xaxis.axis_label = 'px [MeV/c]'
                phase_plot.yaxis.axis_label = 'py [MeV/c]'
                phase_source.data = {'x': df['px'], 'y': df['py']}
        if tabs.active == 1:
            if line_button.active == 0:
                line_plot.xaxis.axis_label = 'z [m]'
                line_plot.yaxis.axis_label = 'rms x [m]'
                line_source.data = {'x': df['z'], 'y': df['rms x']}
            if line_button.active == 1:
                line_plot.xaxis.axis_label = 'z [m]'
                line_plot.yaxis.axis_label = 'rms y [m]'
                line_source.data = {'x': df['z'], 'y': df['rms y']}
            if line_button.active == 2:
                line_plot.xaxis.axis_label = 'z [m]'
                line_plot.yaxis.axis_label = 'rms x Emittance [m rad]'
                line_source.data = {'x': df['z'], 'y': df['rms x Emittance']}
            if line_button.active == 3:
                line_plot.xaxis.axis_label = 'z [m]'
                line_plot.yaxis.axis_label = 'rms y  Emittance [m rad]'
                line_source.data = {'x': df['z'], 'y': df['rms y Emittance']}
            if line_button.active == 4:
                line_plot.xaxis.axis_label = 'z [m]'
                line_plot.yaxis.axis_label = 'avg Energy [MeV]'
                line_source.data = {'x': df['z'], 'y': df['avg Energy']}
            if line_button.active == 5:
                line_plot.xaxis.axis_label = 'z [m]'
                line_plot.yaxis.axis_label = 'x β-function [m]'
                line_source.data = {'x': df['z'], 'y': df['x β-function']}
            if line_button.active == 6:
                line_plot.xaxis.axis_label = 'z [m]'
                line_plot.yaxis.axis_label = 'y β-function [m]'
                line_source.data = {'x': df['z'], 'y': df['y β-function']}
            if line_button.active == 7:
                line_plot.xaxis.axis_label = 'z [m]'
                line_plot.yaxis.axis_label = 'Centroid x [m]'
                line_source.data = {'x': df['z'], 'y': df['centroid x']}
            if line_button.active == 8:
                line_plot.xaxis.axis_label = 'z [m]'
                line_plot.yaxis.axis_label = 'Centroid y [m]'
                line_source.data = {'x': df['z'], 'y': df['centroid y']}
        if tabs.active == 2:
            if field_button.active == 0:
                field_plot.xaxis.axis_label = 'z [m]'
                field_plot.yaxis.axis_label = 'Ez [MV/m]'
                field_source.data = {'x': df['z'], 'y': df['avg Ez']}
            if field_button.active == 1:
                field_plot.xaxis.axis_label = 'z [m]'
                field_plot.yaxis.axis_label = 'Ex [MV/m]'
                field_source.data = {'x': df['z'], 'y': df['avg Ex']}
            if field_button.active == 2:
                field_plot.xaxis.axis_label = 'z [m]'
                field_plot.yaxis.axis_label = 'Ey [MV/m]'
                field_source.data = {'x': df['z'], 'y': df['avg Ey']}
            if field_button.active == 3:
                field_plot.xaxis.axis_label = 'z [m]'
                field_plot.yaxis.axis_label = 'Bz [T]'
                field_source.data = {'x': df['z'], 'y': df['avg Bz']}
            if field_button.active == 4:
                field_plot.xaxis.axis_label = 'z [m]'
                field_plot.yaxis.axis_label = 'Bx [T]'
                field_source.data = {'x': df['z'], 'y': df['avg Bx']}
            if field_button.active == 5:
                field_plot.xaxis.axis_label = 'z [m]'
                field_plot.yaxis.axis_label = 'By [T]'
                field_source.data = {'x': df['z'], 'y': df['avg By']}

    track_button.on_click(track_handler)
    refresh_button.on_click(refresh_handler)

    return tab



import redpic as rp
import pandas as pd
import numpy as np
from bokeh.models import (Panel, PreText, RadioButtonGroup, Select,
                           FileInput, Button, TextInput, ColumnDataSource)
from bokeh.plotting import figure
from bokeh.layouts import column, row
from bokeh.models.widgets import Slider
from os.path import dirname, join

def beam_tab(beam):
    '''Creating a beam with parameters

    '''

    pre = PreText(text='''Here you can generate or upload a beam.''')

    species_button = RadioButtonGroup(
        labels=['electron', 'positron', 'proton', 'antiproton'], active=0)

    select_quantity = Slider(start=1_000, end=100_000,
                             step=1_000, value=1_000,
                             title='Number of particles')

    select_current = Slider(start=0, end=5_000,
                            step=100, value=0,
                            title='Curent [A]')
    # Phase ellipse
    x_input = TextInput(value="0", title="X [m]:", width=75)
    y_input = TextInput(value="0", title="Y [m]:", width=75)
    z_input = TextInput(value="0", title="Z [m]:", width=75)

    px_input = TextInput(value="0", title="Px [MeV/c]:", width=75)
    py_input = TextInput(value="0", title="Py [MeV/c]:", width=75)
    pz_input = TextInput(value="0", title="Pz [MeV/c]:", width=75)

    x_off_input = TextInput(value="0", title="X_off [m]:", width=75)
    y_off_input = TextInput(value="0", title="Y_off [m]:", width=75)
    px_off_input = TextInput(value="0", title="Px_off:", width=60)
    py_off_input = TextInput(value="0", title="Py_off:", width=60)

    phase_ell = column(row(x_input, y_input, z_input, x_off_input),
                       row(px_input, py_input, pz_input, y_off_input))

    select_distribution = RadioButtonGroup(labels=['Uniform', 'Gauss'],
                                           active=0)

    file_input = FileInput(accept='.csv, .ini')
    generate_button = Button(label='Generate', button_type="success",
                             width=155)
    upload_button = Button(label='Upload', button_type="warning",
                           width=155)

    source = ColumnDataSource(data={'x': beam.df['x'], 'y': beam.df['y']})
    p = figure(x_axis_label='x [m]', y_axis_label='y [m]',
               width=410, height=400)
    p.scatter('x', 'y', source=source, size=0.5, color='#3A5785', alpha=0.5)

    controls = column(pre, species_button, select_quantity,
                      select_distribution, select_current,
                      phase_ell, file_input,
                      row(upload_button, generate_button))
    # Create a row layout
    layout = row(controls, p)

    tab = Panel(child=layout, title='Beam')

    def generate_handler(new, beam=beam):
        I = select_current.value
        N = select_quantity.value
        X = float(x_input.value)
        Y = float(y_input.value)
        Z = float(z_input.value)
        Px = float(px_input.value)
        Py = float(py_input.value)
        Pz = float(pz_input.value)
        X_off = float(x_off_input.value)
        Y_off = float(y_off_input.value)
        Px_off = float(px_off_input.value)
        Py_off = float(py_off_input.value)
        if select_distribution.active == 0:
            distribution = rp.Distribution(name='KV', x=X, y=Y, z=Z,
                                           px=Px, py=Py, pz=Pz)
        if select_distribution.active == 1:
            distribution = rp.Distribution(name='GA', x=X, y=Y, z=Z,
                                           px=Px, py=Py, pz=Pz)
        if species_button.active == 0:
            species = rp.electron
        if species_button.active == 1:
            species = rp.positron
        if species_button.active == 2:
            species = rp.proton
        if species_button.active == 3:
            species = rp.antiproton
        beam.type = species
        Q = np.sign(species.charge) * I * Z / rp.c
        beam.generate(distribution, n=N, charge=Q, path=dirname(__file__) + '/data/',
                      x_off=X_off, y_off=Y_off, px_off=Px_off, py_off=Py_off)
        source.data = {'x': beam.df['x'], 'y': beam.df['y']}
        print(beam)

    def upload_handler(new, beam=beam):
        I = select_current.value
        if species_button.active == 0:
            species = rp.electron
        if species_button.active == 1:
            species = rp.positron
        if species_button.active == 2:
            species = rp.proton
        if species_button.active == 3:
            species = rp.antiproton
        beam.type = species
        Q = np.sign(species.charge)* I * (beam.df['z'].max()-beam.df['z'].min()) / rp.c
        beam.upload(dirname(__file__) + '/data/' + file_input.filename, charge=Q,
                    path=dirname(__file__) + '/data/')
        source.data = {'x': beam.df['x'], 'y': beam.df['y']}
        x_input.value = str(np.around(beam.df['x'].max(), 3))
        y_input.value = str(np.around(beam.df['y'].max(), 3))
        z_input.value = str(np.around(beam.df['z'].max(), 3) * 2)
        px_input.value = str(np.around(beam.df['px'].max(), 3))
        py_input.value = str(np.around(beam.df['py'].max(), 3))
        pz_input.value = str(np.around(beam.df['pz'].max(), 3))
        select_quantity.value = beam.n
        print(beam)

    generate_button.on_click(generate_handler)
    upload_button.on_click(upload_handler)

    return tab



'''
Relativictic Difference Scheme Particle-in-Cell code (REDPIC) beam file.

'''

import numpy as np
import pandas as pd
from collections import namedtuple
from .constants import *

__all__ = [ 'Distribution',
            'Beam' ]

def read_distribution_file(fname):
    ''' Read distribution from file .csv

    '''
    cols = ['x', 'y', 'z', 'px', 'py', 'pz']
    #        m    m    m    MeV/c  MeV/c  MeV/c

    df = pd.read_csv(fname, dtype='float32')
    return df['x'], df['y'], df['z'], df['px'], df['py'], df['pz']


def read_distribution_astra(fname):
    ''' Read distribution from Astra file

    '''
    cols = ['x', 'y', 'z', 'px', 'py', 'pz', 'clock', 'charge', 'id', 'flag']
    #        m    m    m    eV/c  eV/c  eV/c  ns       nC
    df = pd.read_csv(fname, header=None, delim_whitespace=True, names=cols, dtype='float32')

    df = df[df.flag != -15] # ignore the lost particles
    df['px'] = df['px']/1e6 # MeV/c
    df['py'] = df['py']/1e6 # MeV/c

    # remove the reference particle
    df0 = df.head(1)
    df  = df.drop(df0.index)

    z0  = df0.z.values[0]
    pz0 = df0.pz.values[0]

    # Recalculate z and pz:
    df['z'] = z0 + df['clock']*1e-9*c # m
    df['pz'] = (pz0 + df['pz'])/1e6 # MeV/c
    return df['x'], df['y'], df['z'], df['px'], df['py'], df['pz']

Distribution = namedtuple('Distribution', [ 'name', 'x', 'y', 'z', 'px', 'py', 'pz' ])

class Beam:
    '''Creates a beam of selected type particles

    '''

    def __init__(self, type: Element) -> None:
        self.type = type        # particles type
        self.charge = 0.0    # beam charge
        self.n = 0.0            # quantity
        self.da = np.array      # data array
        self.df = pd.DataFrame  # data frame

    def generate(self, distribution: Distribution, *, n: float, charge: float=0.0,
                 x_off: float=0.0, y_off: float=0.0, z_off: float=0.0,
                 px_off: float=0.0, py_off: float=0.0, sig_pz: float=0.001,
                 path: str='') -> None:
        '''Beam generator

        This function generates a beam with a given distribution and initial beam displacement.
        '''
        condition = ((distribution.name == 'KV' or distribution.name == 'Uniform') or
                    (distribution.name == 'Gauss' or distribution.name == 'GA'))

        assert n > 0, 'The number of particles (n) must be a positive number!'
        assert type(distribution.name) == str, 'Distribution name must be a string!'
        assert condition, 'Сheck distribution name!'

        self.n = n
        self.charge = charge

        if distribution.name == 'Uniform' or distribution.name == 'KV':
            s = np.random.normal(0, 1, int(self.n))
            t = np.random.normal(0, 1, int(self.n))
            u = np.random.normal(0, 1, int(self.n))
            v = np.random.normal(0, 1, int(self.n))
            norm = (s*s + t*t + u*u + v*v)**0.5
            (x,y) = (distribution.x*u/norm+x_off, distribution.y*v/norm+y_off)
            (px,py) = (distribution.px*u/norm+px_off, distribution.py*v/norm+py_off)
        if distribution.name == 'Gauss' or distribution.name == 'GA':
            x = np.random.normal(x_off, distribution.x, int(self.n))
            y = np.random.normal(y_off, distribution.y, int(self.n))
            px = np.random.normal(px_off, distribution.px, int(self.n))
            py = np.random.normal(py_off, distribution.py, int(self.n))

        z = np.random.uniform(-distribution.z+z_off, distribution.z+z_off, int(self.n)) / 2
        pz = np.random.normal(distribution.pz, distribution.pz * sig_pz, int(self.n))

        self.da = np.row_stack((x, y, z, px, py, pz))
        self.df = pd.DataFrame(np.transpose(self.da), columns=['x','y','z','px','py','pz'])
        self.df.to_csv(path + self.type.symbol + 'Beam.csv')

    def upload(self, file_name: str, *, charge: float=0.0, path: str=''):
        ''' Particle loading

        '''
        file_extension = file_name.split('.')[-1]
        self.charge = charge

        if file_extension == 'ini':
            x, y, z, px, py, pz = read_distribution_astra(file_name)
            self.n = int(len(x))
            self.da = np.row_stack((x, y, z, px, py, pz))
            self.df = pd.DataFrame(np.transpose(self.da), columns=['x','y','z','px','py','pz'])
            self.df.to_csv(path + self.type.symbol + 'Beam.csv')
        if file_extension == 'csv':
            x, y, z, px, py, pz = read_distribution_file(file_name)
            self.n = int(len(x))
            self.da = np.row_stack((x, y, z, px, py, pz))
            self.df = pd.DataFrame(np.transpose(self.da), columns=['x','y','z','px','py','pz'])
            self.df.to_csv(path + self.type.symbol + 'Beam.csv')

    def __str__(self):
        return str(self.df)

    def _ipython_display_(self):
        print(str(self.df))
